##############################################################################
#
# Copyright (c) 2001-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from .common import BTreeTests
from .common import ExtendedSetTests
from .common import I_SetsBase
from .common import InternalKeysMappingTest
from .common import InternalKeysSetTest
from .common import MappingBase
from .common import MappingConflictTestBase
from .common import ModuleTest
from .common import MultiUnion
from .common import NormalSetTests
from .common import SetConflictTestBase
from .common import SetResult
from .common import TestLongIntKeys
from .common import TestLongIntValues
from .common import Weighted
from .common import itemsToSet
from .common import makeBuilder
from BTrees.IIBTree import using64bits #XXX Ugly, but unavoidable

 
class IIBTreeInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBTree
        return IIBTree

 
class IIBTreePyInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBTreePy
        return IIBTreePy


class IITreeSetInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSet
        return IITreeSet


class IITreeSetPyInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSetPy
        return IITreeSetPy


class IIBucketTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBucket
        return IIBucket


class IIBucketPyTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBucketPy
        return IIBucketPy


class IITreeSetTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSet
        return IITreeSet


class IITreeSetPyTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSetPy
        return IITreeSetPy


class IISetTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IISet
        return IISet


class IISetPyTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IISetPy
        return IISetPy


class _IIBTreeTestBase(BTreeTests):

    def testIIBTreeOverflow(self):
        good = set()
        b = self._makeOne()

        def trial(i):
            i = int(i)
            try:
                b[i] = 0
            except OverflowError:
                self.assertRaises(OverflowError, b.__setitem__, 0, i)
            except TypeError:
                self.assertRaises(TypeError, b.__setitem__, 0, i)
            else:
                good.add(i)
                b[0] = i
                self.assertEqual(b[0], i)

        for i in range((1<<31) - 3, (1<<31) + 3):
            trial(i)
            trial(-i)

        del b[0]
        self.assertEqual(sorted(good), sorted(b))


class IIBTreeTest(_IIBTreeTestBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IIBTree
        return IIBTree()


class IIBTreeTestPy(_IIBTreeTestBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IIBTreePy
        return IIBTreePy()


if using64bits:

    class IIBTreeTest(BTreeTests, TestLongIntKeys, TestLongIntValues,
                      unittest.TestCase):

        def _makeOne(self):
            from BTrees.IIBTree import IIBTree
            return IIBTree()

        def getTwoValues(self):
            return 1, 2

    class IIBTreeTest(BTreeTests, TestLongIntKeys, TestLongIntValues,
                      unittest.TestCase):

        def _makeOne(self):
            from BTrees.IIBTree import IIBTreePy
            return IIBTreePy()

        def getTwoValues(self):
            return 1, 2


class _TestIIBTreesBase(object):

    def testNonIntegerKeyRaises(self):
        self.assertRaises(TypeError, self._stringraiseskey)
        self.assertRaises(TypeError, self._floatraiseskey)
        self.assertRaises(TypeError, self._noneraiseskey)

    def testNonIntegerValueRaises(self):
        self.assertRaises(TypeError, self._stringraisesvalue)
        self.assertRaises(TypeError, self._floatraisesvalue)
        self.assertRaises(TypeError, self._noneraisesvalue)

    def _stringraiseskey(self):
        self._makeOne()['c'] = 1

    def _floatraiseskey(self):
        self._makeOne()[2.5] = 1

    def _noneraiseskey(self):
        self._makeOne()[None] = 1

    def _stringraisesvalue(self):
        self._makeOne()[1] = 'c'

    def _floatraisesvalue(self):
        self._makeOne()[1] = 1.4

    def _noneraisesvalue(self):
        self._makeOne()[1] = None


class TestIIBTrees(_TestIIBTreesBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IIBTree
        return IIBTree()


class TestIIBTreesPy(_TestIIBTreesBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IIBTreePy
        return IIBTreePy()


class TestIISets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IISet
        return IISet()


class TestIISetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IISetPy
        return IISetPy()


class TestIITreeSets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IITreeSet
        return IITreeSet()


class TestIITreeSetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IIBTree import IITreeSetPy
        return IITreeSetPy()


class PureII(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.IIBTree import union
        return union(*args)

    def intersection(self, *args):
        from BTrees.IIBTree import intersection
        return intersection(*args)

    def difference(self, *args):
        from BTrees.IIBTree import difference
        return difference(*args)

    def builders(self):
        from BTrees.IIBTree import IIBTree
        from BTrees.IIBTree import IIBucket
        from BTrees.IIBTree import IITreeSet
        from BTrees.IIBTree import IISet
        return IISet, IITreeSet, makeBuilder(IIBTree), makeBuilder(IIBucket)


class PureIIPy(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.IIBTree import unionPy
        return unionPy(*args)

    def intersection(self, *args):
        from BTrees.IIBTree import intersectionPy
        return intersectionPy(*args)

    def difference(self, *args):
        from BTrees.IIBTree import differencePy
        return differencePy(*args)

    def builders(self):
        from BTrees.IIBTree import IIBTreePy
        from BTrees.IIBTree import IIBucketPy
        from BTrees.IIBTree import IITreeSetPy
        from BTrees.IIBTree import IISetPy
        return (IISetPy, IITreeSetPy,
                makeBuilder(IIBTreePy), makeBuilder(IIBucketPy))


class TestIIMultiUnion(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.IIBTree import multiunion
        return multiunion(*args)

    def union(self, *args):
        from BTrees.IIBTree import union
        return union(*args)

    def mkset(self, *args):
        from BTrees.IIBTree import IISet as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.IIBTree import IITreeSet as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.IIBTree import IIBucket as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.IIBTree import IIBTree as mkbtree
        return mkbtree(*args)


class TestIIMultiUnionPy(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.IIBTree import multiunionPy
        return multiunionPy(*args)

    def union(self, *args):
        from BTrees.IIBTree import unionPy
        return unionPy(*args)

    def mkset(self, *args):
        from BTrees.IIBTree import IISetPy as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.IIBTree import IITreeSetPy as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.IIBTree import IIBucketPy as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.IIBTree import IIBTreePy as mkbtree
        return mkbtree(*args)


class TestWeightedII(Weighted, unittest.TestCase):

    def weightedUnion(self):
        from BTrees.IIBTree import weightedUnion
        return weightedUnion

    def weightedIntersection(self):
        from BTrees.IIBTree import weightedIntersection
        return weightedIntersection

    def union(self):
        from BTrees.IIBTree import union
        return union

    def intersection(self):
        from BTrees.IIBTree import intersection
        return intersection

    def mkbucket(self, *args):
        from BTrees.IIBTree import IIBucket as mkbucket
        return mkbucket(*args)

    def builders(self):
        from BTrees.IIBTree import IIBTree
        from BTrees.IIBTree import IIBucket
        from BTrees.IIBTree import IITreeSet
        from BTrees.IIBTree import IISet
        return IIBucket, IIBTree, itemsToSet(IISet), itemsToSet(IITreeSet)


class TestWeightedIIPy(Weighted, unittest.TestCase):

    def weightedUnion(self):
        from BTrees.IIBTree import weightedUnionPy
        return weightedUnionPy

    def weightedIntersection(self):
        from BTrees.IIBTree import weightedIntersectionPy
        return weightedIntersectionPy

    def union(self):
        from BTrees.IIBTree import unionPy
        return unionPy

    def intersection(self):
        from BTrees.IIBTree import intersectionPy
        return intersectionPy

    def mkbucket(self, *args):
        from BTrees.IIBTree import IIBucketPy as mkbucket
        return mkbucket(*args)

    def builders(self):
        from BTrees.IIBTree import IIBTreePy
        from BTrees.IIBTree import IIBucketPy
        from BTrees.IIBTree import IITreeSetPy
        from BTrees.IIBTree import IISetPy
        return (IIBucketPy, IIBTreePy,
                itemsToSet(IISetPy), itemsToSet(IITreeSetPy))


class IIBTreeConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBTree
        return IIBTree


class IIBTreeConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBTreePy
        return IIBTreePy


class IIBucketConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBucket
        return IIBucket


class IIBucketConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IIBucketPy
        return IIBucketPy


class IITreeSetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSet
        return IITreeSet


class IITreeSetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IITreeSetPy
        return IITreeSetPy


class IISetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IISet
        return IISet


class IISetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IIBTree import IISetPy
        return IISetPy


class IIModuleTest(ModuleTest, unittest.TestCase):

    prefix = 'II'

    def _getModule(self):
        import BTrees
        return BTrees.IIBTree
    def _getInterface(self):
        import BTrees.Interfaces
        return BTrees.Interfaces.IIntegerIntegerBTreeModule



def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(IIBTreeInternalKeyTest),
        unittest.makeSuite(IIBTreePyInternalKeyTest),
        unittest.makeSuite(IITreeSetInternalKeyTest),
        unittest.makeSuite(IITreeSetPyInternalKeyTest),
        unittest.makeSuite(IIBucketTest),
        unittest.makeSuite(IIBucketPyTest),
        unittest.makeSuite(IITreeSetTest),
        unittest.makeSuite(IITreeSetPyTest),
        unittest.makeSuite(IISetTest),
        unittest.makeSuite(IISetPyTest),
        unittest.makeSuite(IIBTreeTest),
        unittest.makeSuite(IIBTreeTestPy),
        unittest.makeSuite(TestIIBTrees),
        unittest.makeSuite(TestIIBTreesPy),
        unittest.makeSuite(TestIISets),
        unittest.makeSuite(TestIISetsPy),
        unittest.makeSuite(TestIITreeSets),
        unittest.makeSuite(TestIITreeSetsPy),
        unittest.makeSuite(TestIIMultiUnion),
        unittest.makeSuite(TestIIMultiUnionPy),
        unittest.makeSuite(PureII),
        unittest.makeSuite(PureIIPy),
        unittest.makeSuite(TestWeightedII),
        unittest.makeSuite(TestWeightedIIPy),
        unittest.makeSuite(IIBTreeConflictTests),
        unittest.makeSuite(IIBTreeConflictTestsPy),
        unittest.makeSuite(IIBucketConflictTests),
        unittest.makeSuite(IIBucketConflictTestsPy),
        unittest.makeSuite(IITreeSetConflictTests),
        unittest.makeSuite(IITreeSetConflictTestsPy),
        unittest.makeSuite(IISetConflictTests),
        unittest.makeSuite(IISetConflictTestsPy),
        unittest.makeSuite(IIModuleTest),
    ))



##############################################################################
#
# Copyright (c) 2001-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

__all__ = ('Bucket', 'Set', 'BTree', 'TreeSet',
           'OOBucket', 'OOSet', 'OOBTree', 'OOTreeSet',
           'union', 'intersection','difference',
          )

from zope.interface import moduleProvides

from .Interfaces import IObjectObjectBTreeModule
from ._base import Bucket
from ._base import Set
from ._base import Tree as BTree
from ._base import TreeSet
from ._base import _TreeIterator
from ._base import difference as _difference
from ._base import intersection as _intersection
from ._base import set_operation as _set_operation
from ._base import to_ob as _to_key
_to_value = _to_key
from ._base import union as _union
from ._base import _fix_pickle
from ._compat import import_c_extension

_BUCKET_SIZE = 30
_TREE_SIZE = 250
using64bits = False


class OOBucketPy(Bucket):
    _to_key = _to_key
    _to_value = _to_value


class OOSetPy(Set):
    _to_key = _to_key


class OOBTreePy(BTree):
    max_leaf_size = _BUCKET_SIZE
    max_internal_size = _TREE_SIZE
    _to_key = _to_key
    _to_value = _to_value


class OOTreeSetPy(TreeSet):
    max_leaf_size = _BUCKET_SIZE
    max_internal_size = _TREE_SIZE
    _to_key = _to_key


class OOTreeIteratorPy(_TreeIterator):
    pass


# Can't declare forward refs, so fix up afterwards:

OOBucketPy._mapping_type = OOBucketPy._bucket_type = OOBucketPy
OOBucketPy._set_type = OOSetPy

OOSetPy._mapping_type = OOBucketPy
OOSetPy._set_type = OOSetPy._bucket_type = OOSetPy

OOBTreePy._mapping_type = OOBTreePy._bucket_type = OOBucketPy
OOBTreePy._set_type = OOSetPy

OOTreeSetPy._mapping_type = OOBucketPy
OOTreeSetPy._set_type = OOTreeSetPy._bucket_type = OOSetPy


differencePy = _set_operation(_difference, OOSetPy)
unionPy = _set_operation(_union, OOSetPy)
intersectionPy = _set_operation(_intersection, OOSetPy)

import_c_extension(globals())

_fix_pickle(globals(), __name__)

moduleProvides(IObjectObjectBTreeModule)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from .common import _skip_wo_ZODB
from .common import ConflictTestBase


class NastyConfictFunctionalTests(ConflictTestBase, unittest.TestCase):
    # FUNCTESTS: Provoke various conflict scenarios using ZODB + transaction

    def _getTargetClass(self):
        from BTrees.OOBTree import OOBTree
        return OOBTree

    def openDB(self):
        # The conflict tests tend to open two or more connections
        # and then try to commit them. A standard FileStorage
        # is not MVCC aware, and so each connection would have the same
        # instance of the storage, leading to the error
        # "Duplicate tpc_begin calls for same transaction" on commit;
        # thus we use a MVCCMappingStorage for these tests, ensuring each
        # connection has its own storage.
        # Unfortunately, it wants to acquire the identically same
        # non-recursive lock in each of its *its* tpc_* methods, which deadlocks.
        # The solution is to give each instance its own lock, and trust in the
        # serialization (ordering) of the datamanager, and the fact that these tests are
        # single-threaded.
        import threading
        from ZODB.tests.MVCCMappingStorage  import MVCCMappingStorage
        class _MVCCMappingStorage(MVCCMappingStorage):
            def new_instance(self):
                inst = MVCCMappingStorage.new_instance(self)
                inst._commit_lock = threading.Lock()
                return inst
        from ZODB.DB import DB
        self.storage = _MVCCMappingStorage()
        self.db = DB(self.storage)
        return self.db


    @_skip_wo_ZODB
    def testSimpleConflict(self):
        # Invoke conflict resolution by committing a transaction and
        # catching a conflict in the storage.
        import transaction
        self.openDB()

        r1 = self.db.open().root()
        r1["t"] = t = self._makeOne()
        transaction.commit()

        r2 = self.db.open().root()
        copy = r2["t"]
        list(copy)    # unghostify

        self.assertEqual(t._p_serial, copy._p_serial)

        t.update({1:2, 2:3})
        transaction.commit()

        copy.update({3:4})
        transaction.commit()

    # This tests a problem that cropped up while trying to write
    # testBucketSplitConflict (below):  conflict resolution wasn't
    # working at all in non-trivial cases.  Symptoms varied from
    # strange complaints about pickling (despite that the test isn't
    # doing any *directly*), thru SystemErrors from Python and
    # AssertionErrors inside the BTree code.
    @_skip_wo_ZODB
    def testResolutionBlowsUp(self):
        import transaction
        b = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Invoke conflict resolution by committing a transaction.
        self.openDB()

        r1 = self.db.open().root()
        r1["t"] = b
        transaction.commit()

        r2 = self.db.open().root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(b._p_serial, copy._p_serial)

        b.update({1:2, 2:3})
        transaction.commit()

        copy.update({3:4})
        transaction.commit()  # if this doesn't blow up
        list(copy.values())         # and this doesn't either, then fine

    @_skip_wo_ZODB
    def testBucketSplitConflict(self):
        # Tests that a bucket split is viewed as a conflict.
        # It's (almost necessarily) a white-box test, and sensitive to
        # implementation details.
        import transaction
        from ZODB.POSException import ConflictError
        b = orig = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Invoke conflict resolution by committing a transaction.
        self.openDB()

        tm1 = transaction.TransactionManager()
        r1 = self.db.open(transaction_manager=tm1).root()
        r1["t"] = b
        tm1.commit()

        tm2 = transaction.TransactionManager()
        r2 = self.db.open(transaction_manager=tm2).root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(orig._p_serial, copy._p_serial)

        # In one transaction, add 16 new keys to bucket1, to force a bucket
        # split.
        b = orig
        numtoadd = 16
        candidate = 60
        while numtoadd:
            if candidate not in b:
                b[candidate] = candidate
                numtoadd -= 1
            candidate += 1
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 61 .. 74
        # bucket 2 has 16 values: [75, 76 .. 81] + [84, 88 ..116]
        # bucket 3 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((b0, 60, b1, 75, b2, 120, b3), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state) , 2)
        self.assertEqual(len(state[0]), 7)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 75)
        self.assertEqual(state[0][5], 120)

        tm1.commit()

        # In the other transaction, add 3 values near the tail end of bucket1.
        # This doesn't cause a split.
        b = copy
        for i in range(112, 116):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 18 values: 60, 64 .. 112, 113, 114, 115, 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        self.assertRaises(ConflictError, tm2.commit)

    @_skip_wo_ZODB
    def testEmptyBucketConflict(self):
        # Tests that an emptied bucket *created by* conflict resolution is
        # viewed as a conflict:  conflict resolution doesn't have enough
        # info to unlink the empty bucket from the BTree correctly.
        import transaction
        from ZODB.POSException import ConflictError
        b = orig = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Invoke conflict resolution by committing a transaction.
        self.openDB()

        tm1 = transaction.TransactionManager()
        r1 = self.db.open(transaction_manager=tm1).root()
        r1["t"] = b
        tm1.commit()

        tm2 = transaction.TransactionManager()
        r2 = self.db.open(transaction_manager=tm2).root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(orig._p_serial, copy._p_serial)

        # In one transaction, delete half of bucket 1.
        b = orig
        for k in 60, 64, 68, 72, 76, 80, 84, 88:
            del b[k]
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 7 values: 92, 96, 100, 104, 108, 112, 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state) , 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 92)
        self.assertEqual(state[0][3], 120)

        tm1.commit()

        # In the other transaction, delete the other half of bucket 1.
        b = copy
        for k in 92, 96, 100, 104, 108, 112, 116:
            del b[k]
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 8 values: 60, 64, 68, 72, 76, 80, 84, 88
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Conflict resolution empties bucket1 entirely.  This used to
        # create an "insane" BTree (a legit BTree cannot contain an empty
        # bucket -- it contains NULL pointers the BTree code doesn't
        # expect, and segfaults result).
        self.assertRaises(ConflictError, tm2.commit)

    @_skip_wo_ZODB
    def testEmptyBucketNoConflict(self):
        # Tests that a plain empty bucket (on input) is not viewed as a
        # conflict.
        import transaction
        b = orig = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Invoke conflict resolution by committing a transaction.
        self.openDB()

        r1 = self.db.open().root()
        r1["t"] = orig
        transaction.commit()

        r2 = self.db.open().root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(orig._p_serial, copy._p_serial)

        # In one transaction, just add a key.
        b = orig
        b[1] = 1
        # bucket 0 has 16 values: [0, 1] + [4, 8 .. 56]
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        transaction.commit()

        # In the other transaction, delete bucket 2.
        b = copy
        for k in range(120, 200, 4):
            del b[k]
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1), firstbucket)
        # The next block is still verifying preconditions.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 3)
        self.assertEqual(state[0][1], 60)

        # This shouldn't create a ConflictError.
        transaction.commit()
        # And the resulting BTree shouldn't have internal damage.
        b._check()

    # The snaky control flow in _bucket__p_resolveConflict ended up trying
    # to decref a NULL pointer if conflict resolution was fed 3 empty
    # buckets.  http://collector.zope.org/Zope/553
    def testThreeEmptyBucketsNoSegfault(self):
        # Note that the conflict is raised by our C extension, rather than
        # indirectly via the storage, and hence is a more specialized type.
        # This test therefore does not require ZODB.
        from BTrees.Interfaces import BTreesConflictError
        t = self._makeOne()
        t[1] = 1
        bucket = t._firstbucket
        del t[1]
        state1 = bucket.__getstate__()
        state2 = bucket.__getstate__()
        state3 = bucket.__getstate__()
        self.assertTrue(state2 is not state1 and
                        state2 is not state3 and
                        state3 is not state1)
        self.assertTrue(state2 == state1 and
                        state3 == state1)
        self.assertRaises(BTreesConflictError, bucket._p_resolveConflict,
                          state1, state2, state3)
        # When an empty BTree resolves conflicts, it computes the
        # bucket state as None, so...
        self.assertRaises(BTreesConflictError, bucket._p_resolveConflict,
                          None, None, None)

    @_skip_wo_ZODB
    def testCantResolveBTreeConflict(self):
        # Test that a conflict involving two different changes to
        # an internal BTree node is unresolvable.  An internal node
        # only changes when there are enough additions or deletions
        # to a child bucket that the bucket is split or removed.
        # It's (almost necessarily) a white-box test, and sensitive to
        # implementation details.
        import transaction
        from ZODB.POSException import ConflictError
        b = orig = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Set up database connections to provoke conflict.
        self.openDB()
        tm1 = transaction.TransactionManager()
        r1 = self.db.open(transaction_manager=tm1).root()
        r1["t"] = orig
        tm1.commit()

        tm2 = transaction.TransactionManager()
        r2 = self.db.open(transaction_manager=tm2).root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(orig._p_serial, copy._p_serial)

        # Now one transaction should add enough keys to cause a split,
        # and another should remove all the keys in one bucket.

        for k in range(200, 300, 4):
            orig[k] = k
        tm1.commit()

        for k in range(0, 60, 4):
            del copy[k]

        self.assertRaises(ConflictError, tm2.commit)

    @_skip_wo_ZODB
    def testConflictWithOneEmptyBucket(self):
        # If one transaction empties a bucket, while another adds an item
        # to the bucket, all the changes "look resolvable":  bucket conflict
        # resolution returns a bucket containing (only) the item added by
        # the latter transaction, but changes from the former transaction
        # removing the bucket are uncontested:  the bucket is removed from
        # the BTree despite that resolution thinks it's non-empty!  This
        # was first reported by Dieter Maurer, to zodb-dev on 22 Mar 2005.
        import transaction
        from ZODB.POSException import ConflictError
        b = orig = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i
        # bucket 0 has 15 values: 0, 4 .. 56
        # bucket 1 has 15 values: 60, 64 .. 116
        # bucket 2 has 20 values: 120, 124 .. 196
        state = b.__getstate__()
        # Looks like:  ((bucket0, 60, bucket1, 120, bucket2), firstbucket)
        # If these fail, the *preconditions* for running the test aren't
        # satisfied -- the test itself hasn't been run yet.
        self.assertEqual(len(state), 2)
        self.assertEqual(len(state[0]), 5)
        self.assertEqual(state[0][1], 60)
        self.assertEqual(state[0][3], 120)

        # Set up database connections to provoke conflict.
        self.openDB()
        tm1 = transaction.TransactionManager()
        r1 = self.db.open(transaction_manager=tm1).root()
        r1["t"] = orig
        tm1.commit()

        tm2 = transaction.TransactionManager()
        r2 = self.db.open(transaction_manager=tm2).root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(orig._p_serial, copy._p_serial)

        # Now one transaction empties the first bucket, and another adds a
        # key to the first bucket.

        for k in range(0, 60, 4):
            del orig[k]
        tm1.commit()

        copy[1] = 1

        self.assertRaises(ConflictError, tm2.commit)

        # Same thing, except commit the transactions in the opposite order.
        b = self._makeOne()
        for i in range(0, 200, 4):
            b[i] = i

        tm1 = transaction.TransactionManager()
        r1 = self.db.open(transaction_manager=tm1).root()
        r1["t"] = b
        tm1.commit()

        tm2 = transaction.TransactionManager()
        r2 = self.db.open(transaction_manager=tm2).root()
        copy = r2["t"]
        # Make sure all of copy is loaded.
        list(copy.values())

        self.assertEqual(b._p_serial, copy._p_serial)

        # Now one transaction empties the first bucket, and another adds a
        # key to the first bucket.
        b[1] = 1
        tm1.commit()

        for k in range(0, 60, 4):
            del copy[k]

        self.assertRaises(ConflictError, tm2.commit)

    @_skip_wo_ZODB
    def testConflictOfInsertAndDeleteOfFirstBucketItem(self):
        # Recently, BTrees became careful about removing internal keys
        # (keys in internal aka BTree nodes) when they were deleted from
        # buckets. This poses a problem for conflict resolution.

        # We want to guard against a case in which the first key in a
        # bucket is removed in one transaction while a key is added
        # after that key but before the next key in another transaction
        # with the result that the added key is unreachable.

        # original:

        #   Bucket(...), k1, Bucket((k1, v1), (k3, v3), ...)

        # tran1

        #   Bucket(...), k3, Bucket(k3, v3), ...)

        # tran2

        #   Bucket(...), k1, Bucket((k1, v1), (k2, v2), (k3, v3), ...)

        #   where k1 < k2 < k3

        # We don't want:

        #   Bucket(...), k3, Bucket((k2, v2), (k3, v3), ...)

        #   as k2 would be unfindable, so we want a conflict.

        import transaction
        from ZODB.POSException import ConflictError
        mytype = self._getTargetClass()
        db = self.openDB()
        tm1 = transaction.TransactionManager()
        conn1 = db.open(tm1)
        conn1.root.t = t = mytype()
        for i in range(0, 200, 2):
            t[i] = i
        tm1.commit()
        k = t.__getstate__()[0][1]
        assert t.__getstate__()[0][2].keys()[0] == k

        tm2 = transaction.TransactionManager()
        conn2 = db.open(tm2)

        t[k+1] = k+1
        del conn2.root.t[k]
        for i in range(200,300):
            conn2.root.t[i] = i

        tm1.commit()
        self.assertRaises(ConflictError, tm2.commit)
        tm2.abort()

        k = t.__getstate__()[0][1]
        t[k+1] = k+1
        del conn2.root.t[k]

        tm2.commit()
        self.assertRaises(ConflictError, tm1.commit)
        tm1.abort()


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(NastyConfictFunctionalTests),
    ))



##############################################################################
#
# Copyright (c) 2001-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from .common import BTreeTests
from .common import ExtendedSetTests
from .common import I_SetsBase
from .common import InternalKeysMappingTest
from .common import InternalKeysSetTest
from .common import MappingBase
from .common import MappingConflictTestBase
from .common import ModuleTest
from .common import MultiUnion
from .common import NormalSetTests
from .common import SetConflictTestBase
from .common import SetResult
from .common import TypeTest
from .common import TestLongIntKeys
from .common import makeBuilder
from BTrees.IIBTree import using64bits #XXX Ugly, but unavoidable


class IOBTreeInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBTree
        return IOBTree


class IOBTreePyInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBTreePy
        return IOBTreePy


class IOTreeSetInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSet
        return IOTreeSet


class IOTreeSetPyInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSetPy
        return IOTreeSetPy


class IOBucketTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBucket
        return IOBucket


class IOBucketPyTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBucketPy
        return IOBucketPy


class IOTreeSetTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSet
        return IOTreeSet


class IOTreeSetPyTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSetPy
        return IOTreeSetPy


class IOSetTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOSet
        return IOSet


class IOSetPyTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOSetPy
        return IOSetPy


class IOBTreeTest(BTreeTests, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOBTree
        return IOBTree()


class IOBTreePyTest(BTreeTests, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOBTreePy
        return IOBTreePy()


if using64bits:


    class IOBTreeTest(BTreeTests, TestLongIntKeys, unittest.TestCase):

        def _makeOne(self):
            from BTrees.IOBTree import IOBTree
            return IOBTree()


    class IOBTreePyTest(BTreeTests, TestLongIntKeys, unittest.TestCase):

        def _makeOne(self):
            from BTrees.IOBTree import IOBTreePy
            return IOBTreePy()


class _TestIOBTreesBase(TypeTest):

    def _stringraises(self):
        self._makeOne()['c'] = 1

    def _floatraises(self):
        self._makeOne()[2.5] = 1

    def _noneraises(self):
        self._makeOne()[None] = 1

    def testStringAllowedInContains(self):
        self.assertFalse('key' in self._makeOne())

    def testStringKeyRaisesKeyErrorWhenMissing(self):
        self.assertRaises(KeyError, self._makeOne().__getitem__, 'key')

    def testStringKeyReturnsDefaultFromGetWhenMissing(self):
        self.assertEqual(self._makeOne().get('key', 42), 42)

class TestIOBTrees(_TestIOBTreesBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOBTree
        return IOBTree()


class TestIOBTreesPy(_TestIOBTreesBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOBTreePy
        return IOBTreePy()


class TestIOSets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOSet
        return IOSet()


class TestIOSetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOSetPy
        return IOSetPy()


class TestIOTreeSets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOTreeSet
        return IOTreeSet()


class TestIOTreeSetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.IOBTree import IOTreeSetPy
        return IOTreeSetPy()


class PureIO(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.IOBTree import union
        return union(*args)

    def intersection(self, *args):
        from BTrees.IOBTree import intersection
        return intersection(*args)

    def difference(self, *args):
        from BTrees.IOBTree import difference
        return difference(*args)

    def builders(self):
        from BTrees.IOBTree import IOBTree
        from BTrees.IOBTree import IOBucket
        from BTrees.IOBTree import IOTreeSet
        from BTrees.IOBTree import IOSet
        return IOSet, IOTreeSet, makeBuilder(IOBTree), makeBuilder(IOBucket)


class PureIOPy(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.IOBTree import unionPy
        return unionPy(*args)

    def intersection(self, *args):
        from BTrees.IOBTree import intersectionPy
        return intersectionPy(*args)

    def difference(self, *args):
        from BTrees.IOBTree import differencePy
        return differencePy(*args)

    def builders(self):
        from BTrees.IOBTree import IOBTreePy
        from BTrees.IOBTree import IOBucketPy
        from BTrees.IOBTree import IOTreeSetPy
        from BTrees.IOBTree import IOSetPy
        return (IOSetPy, IOTreeSetPy,
                makeBuilder(IOBTreePy), makeBuilder(IOBucketPy))


class TestIOMultiUnion(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.IOBTree import multiunion
        return multiunion(*args)

    def union(self, *args):
        from BTrees.IOBTree import union
        return union(*args)

    def mkset(self, *args):
        from BTrees.IOBTree import IOSet as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.IOBTree import IOTreeSet as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.IOBTree import IOBucket as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.IOBTree import IOBTree as mkbtree
        return mkbtree(*args)


class TestIOMultiUnionPy(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.IOBTree import multiunionPy
        return multiunionPy(*args)

    def union(self, *args):
        from BTrees.IOBTree import unionPy
        return unionPy(*args)

    def mkset(self, *args):
        from BTrees.IOBTree import IOSetPy as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.IOBTree import IOTreeSetPy as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.IOBTree import IOBucketPy as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.IOBTree import IOBTreePy as mkbtree
        return mkbtree(*args)


class IOBTreeConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBTree
        return IOBTree


class IOBTreeConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBTreePy
        return IOBTreePy


class IOBucketConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBucket
        return IOBucket


class IOBucketConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOBucketPy
        return IOBucketPy


class IOTreeSetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSet
        return IOTreeSet


class IOTreeSetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOTreeSetPy
        return IOTreeSetPy


class IOSetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOSet
        return IOSet


class IOSetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.IOBTree import IOSetPy
        return IOSetPy


class IOModuleTest(ModuleTest, unittest.TestCase):

    prefix = 'IO'

    def _getModule(self):
        import BTrees
        return BTrees.IOBTree

    def _getInterface(self):
        import BTrees.Interfaces
        return BTrees.Interfaces.IIntegerObjectBTreeModule

    def test_weightedUnion_not_present(self):
        try:
            from BTrees.IOBTree import weightedUnion
        except ImportError:
            pass
        else:
            self.fail("IOBTree shouldn't have weightedUnion")

    def test_weightedIntersection_not_present(self):
        try:
            from BTrees.IOBTree import weightedIntersection
        except ImportError:
            pass
        else:
            self.fail("IOBTree shouldn't have weightedIntersection")


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(IOBTreeInternalKeyTest),
        unittest.makeSuite(IOBTreePyInternalKeyTest),
        unittest.makeSuite(IOTreeSetInternalKeyTest),
        unittest.makeSuite(IOTreeSetPyInternalKeyTest),
        unittest.makeSuite(IOBucketTest),
        unittest.makeSuite(IOBucketPyTest),
        unittest.makeSuite(IOTreeSetTest),
        unittest.makeSuite(IOTreeSetPyTest),
        unittest.makeSuite(IOSetTest),
        unittest.makeSuite(IOSetPyTest),
        unittest.makeSuite(IOBTreeTest),
        unittest.makeSuite(IOBTreePyTest),
        unittest.makeSuite(TestIOBTrees),
        unittest.makeSuite(TestIOBTreesPy),
        unittest.makeSuite(TestIOSets),
        unittest.makeSuite(TestIOSetsPy),
        unittest.makeSuite(TestIOTreeSets),
        unittest.makeSuite(TestIOTreeSetsPy),
        unittest.makeSuite(TestIOMultiUnion),
        unittest.makeSuite(TestIOMultiUnionPy),
        unittest.makeSuite(PureIO),
        unittest.makeSuite(PureIOPy),
        unittest.makeSuite(IOBTreeConflictTests),
        unittest.makeSuite(IOBTreeConflictTestsPy),
        unittest.makeSuite(IOBucketConflictTests),
        unittest.makeSuite(IOBucketConflictTestsPy),
        unittest.makeSuite(IOTreeSetConflictTests),
        unittest.makeSuite(IOTreeSetConflictTestsPy),
        unittest.makeSuite(IOSetConflictTests),
        unittest.makeSuite(IOSetConflictTestsPy),
        unittest.makeSuite(IOModuleTest),
    ))



##############################################################################
#
# Copyright (c) 2001-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from .common import BTreeTests
from .common import ExtendedSetTests
from .common import I_SetsBase
from .common import InternalKeysMappingTest
from .common import InternalKeysSetTest
from .common import MappingBase
from .common import MappingConflictTestBase
from .common import ModuleTest
from .common import MultiUnion
from .common import NormalSetTests
from .common import SetConflictTestBase
from .common import SetResult
from .common import TestLongIntKeys
from .common import makeBuilder


class LOBTreeInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBTree
        return LOBTree


class LOBTreePyInternalKeyTest(InternalKeysMappingTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBTreePy
        return LOBTreePy


class LOTreeSetInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSet
        return LOTreeSet


class LOTreeSetPyInternalKeyTest(InternalKeysSetTest, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSetPy
        return LOTreeSetPy


class LOBucketTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBucket
        return LOBucket


class LOBucketPyTest(MappingBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBucketPy
        return LOBucketPy


class LOTreeSetTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSet
        return LOTreeSet


class LOTreeSetPyTest(NormalSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSetPy
        return LOTreeSetPy


class LOSetTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOSet
        return LOSet


class LOSetPyTest(ExtendedSetTests, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOSetPy
        return LOSetPy


class LOBTreeTest(BTreeTests, TestLongIntKeys, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOBTree
        return LOBTree()


class LOBTreePyTest(BTreeTests, TestLongIntKeys, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOBTreePy
        return LOBTreePy()


class TestLOSets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOSet
        return LOSet()


class TestLOSetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOSetPy
        return LOSetPy()


class TestLOTreeSets(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOTreeSet
        return LOTreeSet()


class TestLOTreeSetsPy(I_SetsBase, unittest.TestCase):

    def _makeOne(self):
        from BTrees.LOBTree import LOTreeSetPy
        return LOTreeSetPy()


class TestLOMultiUnion(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.LOBTree import multiunion
        return multiunion(*args)

    def union(self, *args):
        from BTrees.LOBTree import union
        return union(*args)

    def mkset(self, *args):
        from BTrees.LOBTree import LOSet as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.LOBTree import LOTreeSet as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.LOBTree import LOBucket as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.LOBTree import LOBTree as mkbtree
        return mkbtree(*args)


class TestLOMultiUnionPy(MultiUnion, unittest.TestCase):

    def multiunion(self, *args):
        from BTrees.LOBTree import multiunionPy
        return multiunionPy(*args)

    def union(self, *args):
        from BTrees.LOBTree import unionPy
        return unionPy(*args)

    def mkset(self, *args):
        from BTrees.LOBTree import LOSetPy as mkset
        return mkset(*args)

    def mktreeset(self, *args):
        from BTrees.LOBTree import LOTreeSetPy as mktreeset
        return mktreeset(*args)

    def mkbucket(self, *args):
        from BTrees.LOBTree import LOBucketPy as mkbucket
        return mkbucket(*args)

    def mkbtree(self, *args):
        from BTrees.LOBTree import LOBTreePy as mkbtree
        return mkbtree(*args)


class PureLO(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.LOBTree import union
        return union(*args)

    def intersection(self, *args):
        from BTrees.LOBTree import intersection
        return intersection(*args)

    def difference(self, *args):
        from BTrees.LOBTree import difference
        return difference(*args)

    def builders(self):
        from BTrees.LOBTree import LOBTree
        from BTrees.LOBTree import LOBucket
        from BTrees.LOBTree import LOTreeSet
        from BTrees.LOBTree import LOSet
        return LOSet, LOTreeSet, makeBuilder(LOBTree), makeBuilder(LOBucket)


class PureLOPy(SetResult, unittest.TestCase):

    def union(self, *args):
        from BTrees.LOBTree import unionPy
        return unionPy(*args)

    def intersection(self, *args):
        from BTrees.LOBTree import intersectionPy
        return intersectionPy(*args)

    def difference(self, *args):
        from BTrees.LOBTree import differencePy
        return differencePy(*args)

    def builders(self):
        from BTrees.LOBTree import LOBTreePy
        from BTrees.LOBTree import LOBucketPy
        from BTrees.LOBTree import LOTreeSetPy
        from BTrees.LOBTree import LOSetPy
        return (LOSetPy, LOTreeSetPy,
                makeBuilder(LOBTreePy), makeBuilder(LOBucketPy))


class LOBTreeConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBTree
        return LOBTree


class LOBTreeConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBTreePy
        return LOBTreePy


class LOBucketConflictTests(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBucket
        return LOBucket


class LOBucketConflictTestsPy(MappingConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOBucketPy
        return LOBucketPy


class LOTreeSetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSet
        return LOTreeSet


class LOTreeSetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOTreeSetPy
        return LOTreeSetPy


class LOSetConflictTests(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOSet
        return LOSet


class LOSetConflictTestsPy(SetConflictTestBase, unittest.TestCase):

    def _getTargetClass(self):
        from BTrees.LOBTree import LOSetPy
        return LOSetPy


class LOModuleTest(ModuleTest, unittest.TestCase):

    prefix = 'LO'

    def _getModule(self):
        import BTrees
        return BTrees.LOBTree

    def _getInterface(self):
        import BTrees.Interfaces
        return BTrees.Interfaces.IIntegerObjectBTreeModule

    def test_weightedUnion_not_present(self):
        try:
            from BTrees.LOBTree import weightedUnion
        except ImportError:
            pass
        else:
            self.fail("LOBTree shouldn't have weightedUnion")

    def test_weightedIntersection_not_present(self):
        try:
            from BTrees.LOBTree import weightedIntersection
        except ImportError:
            pass
        else:
            self.fail("LOBTree shouldn't have weightedIntersection")


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(LOBTreeInternalKeyTest),
        unittest.makeSuite(LOBTreePyInternalKeyTest),
        unittest.makeSuite(LOTreeSetInternalKeyTest),
        unittest.makeSuite(LOTreeSetPyInternalKeyTest),
        unittest.makeSuite(LOBucketTest),
        unittest.makeSuite(LOBucketPyTest),
        unittest.makeSuite(LOTreeSetTest),
        unittest.makeSuite(LOTreeSetPyTest),
        unittest.makeSuite(LOSetTest),
        unittest.makeSuite(LOSetPyTest),
        unittest.makeSuite(LOBTreeTest),
        unittest.makeSuite(LOBTreePyTest),
        unittest.makeSuite(TestLOSets),
        unittest.makeSuite(TestLOSetsPy),
        unittest.makeSuite(TestLOTreeSets),
        unittest.makeSuite(TestLOTreeSetsPy),
        unittest.makeSuite(TestLOMultiUnion),
        unittest.makeSuite(TestLOMultiUnionPy),
        unittest.makeSuite(PureLO),
        unittest.makeSuite(PureLOPy),
        unittest.makeSuite(LOBTreeConflictTests),
        unittest.makeSuite(LOBTreeConflictTestsPy),
        unittest.makeSuite(LOBucketConflictTests),
        unittest.makeSuite(LOBucketConflictTestsPy),
        unittest.makeSuite(LOTreeSetConflictTests),
        unittest.makeSuite(LOTreeSetConflictTestsPy),
        unittest.makeSuite(LOSetConflictTests),
        unittest.makeSuite(LOSetConflictTestsPy),
        unittest.makeSuite(LOModuleTest),
    ))
